
/* Linux driver for Disk-On-Chip 2000       */
/* (c) 1999 Machine Vision Holdings, Inc.   */
/* Author: David Woodhouse <dwmw2@mvhi.com> */
/* $Id: doc2000.c,v 1.9 2000/01/19 10:26:04 dwmw2 Exp $ */


/* DOC_PASSIVE_PROBE:
   In order to ensure that the BIOS checksum is correct at boot time, and 
   hence that the onboard BIOS extension gets executed, the DiskOnChip 
   goes into reset mode when it is read sequentially: all registers 
   return 0xff until the chip is woken up again by writing to the 
   DOCControl register. 

   Unfortunately, this means that the probe for the DiskOnChip is unsafe, 
   because one of the first things it does is write to where it thinks 
   the DOCControl register should be - which may well be shared memory 
   for another device. I've had machines which lock up when this is 
   attempted. Hence the possibility to do a passive probe, which will fail 
   to detect a chip in reset mode, but is at least guaranteed not to lock
   the machine.

   If you have this problem, uncomment the following line:
#define DOC_PASSIVE_PROBE
*/


#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/types.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/doc2000.h>

#ifdef __SMP__
#warning The DiskOnChip driver is not SMP safe. I need to add some locking -dw
#endif

#ifndef THIS_MODULE
#ifdef MODULE
#define THIS_MODULE (&__this_module)
#else
#define THIS_MODULE NULL
#endif
#endif

//#define PRERELEASE

unsigned long doc_locations[] = { 
		0xc8000, 0xca000, 0xcc000, 0xce000, 
		0xd0000, 0xd2000, 0xd4000, 0xd6000,
		0xd8000, 0xda000, 0xdc000, 0xde000, 
		0xe0000, 0xe2000, 0xe4000, 0xe6000, 
		0xe8000, 0xea000, 0xec000, 0xee000, 0 };

static int numdocs = 0;


/* DOC_WaitReady: Wait for RDY line to be asserted by the flash chip */

int _DoC_WaitReady (unsigned long docptr)
{
	long c=-1;

	/* Out-of-line routine to wait for chip response */
	while (!(ReadDOC(docptr, CSDNControl) & CSDN_CTRL_FR_B) && --c)
		;
	
	if (c == 0)
		printk(KERN_WARNING "_DoC_WaitReady timed out.");
	
	return (c==0);
}

static inline int DoC_WaitReady(unsigned long docptr) 
{
	/* This is inline, to optimise the common case, where it's ready instantly */
	
	if (!(ReadDOC(docptr, CSDNControl) & CSDN_CTRL_FR_B))
		return _DoC_WaitReady(docptr); /* Call the out-of-line routine to wait */
	
	return 0;
}


/* DoC_Command: Send a flash command to the flash chip */

static inline int DoC_Command(unsigned long docptr, unsigned char command, unsigned char xtraflags)
{
	/* Assert the CLE (Command Latch Enable) line to the flash chip */
 	WriteDOC( CSDN_CTRL_FLASH_IO | xtraflags | CSDN_CTRL_CLE | CSDN_CTRL_CE,
		 docptr, CSDNControl);

	/* Send the command */
	WriteDOC(command, docptr, 2k_CSDN_IO);
 
	/* Lower the CLE line */
	WriteDOC( CSDN_CTRL_FLASH_IO | xtraflags | CSDN_CTRL_CE, docptr, CSDNControl);

	/* Wait for the chip to respond */
	return DoC_WaitReady(docptr);
}

/* DoC_Address: Set the current address for the flash chip */

static inline int DoC_Address (unsigned long docptr, int numbytes, unsigned long ofs,
			       unsigned char xtraflags1, unsigned char xtraflags2)
{
	/* Assert the ALE (Address Latch Enable line to the flash chip */
 	WriteDOC( CSDN_CTRL_FLASH_IO | xtraflags1 | CSDN_CTRL_ALE | CSDN_CTRL_CE,
		 docptr, CSDNControl);

	/* Send the address */
	/* Three cases:
	   numbytes == 1: Send single byte, bits 0-7.
	   numbytes == 2: Send bits 9-16 followed by 17-23
	   numbytes == 3: Send 0-7, 9-16, then 17-23 
	*/
	if (numbytes != 2)
		WriteDOC(ofs & 0xff, docptr, 2k_CSDN_IO);
	
	if (numbytes != 1) {
		WriteDOC((ofs >> 9) & 0xff, docptr, 2k_CSDN_IO);
		WriteDOC((ofs >> 17) & 0xff, docptr, 2k_CSDN_IO);
	}
	/* Lower the ALE line */
	WriteDOC( CSDN_CTRL_FLASH_IO | xtraflags1 | xtraflags2 | CSDN_CTRL_CE, docptr, CSDNControl);
	
	/* Wait for the chip to respond */
	return DoC_WaitReady(docptr);
}

/* DoC_SelectChip: Select a given flash chip within the current floor */

static inline int DoC_SelectChip(unsigned long docptr, int chip)
{
	/* Select the individual flash chip requested */
	WriteDOC( chip, docptr, CSDNDeviceSelect);
	
	/* Wait for it to be ready */
	return DoC_WaitReady(docptr);
}

/* DoC_SelectFloor: Select a given floor (bank of flash chips) */

static inline int DoC_SelectFloor(unsigned long docptr, int floor)
{
	/* Select the floor (bank) of chips required */
	WriteDOC( floor, docptr, FloorSelect);

	/* Wait for the chip to be ready */
	return DoC_WaitReady(docptr);
}
  
/* DoC_IdentChip: Identify a given NAND chip given {floor,chip} */

int DoC_IdentChip(struct DiskOnChip *doc, int floor, int chip)
{
	int mfr, id, chipshift=0;
	char *mfrname=NULL, *idname=NULL;

	/* Page in the required floor/chip */
	DoC_SelectFloor(doc->virtadr, floor);
	DoC_SelectChip(doc->virtadr, chip);

	/* Reset the chip */
	if (DoC_Command(doc->virtadr, NAND_CMD_RESET, CSDN_CTRL_WP)) {
		printk("DoC_Command (reset) for %d,%d returned true\n", floor,chip);
		return 0;
	}
  
	/* Read the NAND chip ID: 1. Send ReadID command */ 
	if(DoC_Command(doc->virtadr, NAND_CMD_READID, CSDN_CTRL_WP)) {
		printk("DoC_Command (ReadID) for %d,%d returned true\n", floor,chip);
		return 0;
	}

	/* Read the NAND chip ID: 2. Send address byte zero 
	 */ 
	DoC_Address(doc->virtadr, 1, 0, CSDN_CTRL_WP, 0);
 
	/* Read the manufacturer and device id codes from the device */
	mfr = ReadDOC(doc->virtadr, 2k_CSDN_IO);
	id = ReadDOC(doc->virtadr, 2k_CSDN_IO);
	
	/* No response - return failure */
	if (mfr == 0xff || mfr == 0)
		return 0;
	
	/* Check it's the same as the first chip we identified. 
	 * M-Systems say that any given DiskOnChip device should only
	 * contain _one_ type of flash part, although that's not a 
	 * hardware restriction. */
	if (doc->mfr) {
		if (doc->mfr == mfr && doc->id == id)
			return 1; /* This is another the same the first */
		else
			printk(KERN_WARNING "Flash chip at floor %d, chip %d is different:\n",
			       floor, chip);
	}
	
	/* Print (and store if first time) the manufacturer and ID codes. */
  
	switch(mfr) {
	case NAND_MFR_TOSHIBA: /* Toshiba */
		mfrname = "Toshiba";
		
		switch(id) { 
		case 0x64:
			idname = "TC5816BDC";
			chipshift = 21;
			break;
			
		case 0x6b:
			idname = "TC5832DC";
			chipshift = 22;
			break;
			
		case 0x73: 
			idname = "TH58V128DC";
			chipshift = 24;
			break;
			
		case 0x75: 
			idname = "TC58256FT/DC";
			chipshift = 25;
			break;
			
		case 0xe5:
			idname = "TC58V32DC";
			chipshift = 22;
			break;
			
		case 0xe6: 
			idname = "TC58V64DC";
			chipshift = 23;
			break;
			
		case 0xea:
			idname = "TC58V16BDC";
			chipshift = 21;
			break;
		}
		break; /* End of Toshiba parts */
		
	case NAND_MFR_SAMSUNG: /* Samsung */
		mfrname = "Samsung";
		
		switch(id) {
		case 0x64:
			idname = "KM29N16000";
			chipshift = 21;
			
		case 0x73:
			idname = "KM29U128T";
			chipshift = 24;
			break;
			
		case 0x75:
			idname = "KM29U256T";
			chipshift = 25;
			break;

		case 0xe3:
			idname = "KM29W32000";
			chipshift = 22;
			break;
			
		case 0xe6:
			idname = "KM29U64000";
			chipshift = 23;
			break;
			
		case 0xea:
			idname = "KM29W16000";
			chipshift = 21;
			break;
		}
		break; /* End of Samsung parts */
	}
	
	/* If we've identified it fully, print the full names */
	if (idname) {
#ifdef PRERELEASE
		printk (KERN_INFO "Flash chip found: %2.2X %2.2X (%s %s)\n", 
			mfr,id,mfrname,idname);
#endif
		/* If this is the first chip, store the id codes */
		if (!doc->mfr) {
			doc->mfr = mfr;
			doc->id = id;
			doc->chipshift = chipshift;
			return 1;
		}
		return 0;
	}

	/* We haven't fully identified the chip. Print as much as we know. */
	if (mfrname)
		printk(KERN_WARNING "Unknown %s flash chip found: %2.2X %2.2X\n", mfrname,
		       id, mfr);
	else
		printk(KERN_WARNING "Unknown flash chip found: %2.2X %2.2X\n", id, mfr);
	
	printk(KERN_WARNING "Please report to David.Woodhouse@mvhi.com\n");
	return 0;
}     

/* DoC_ScanChips: Find all NAND chips present in a DiskOnChip, and identify them */

void DoC_ScanChips(struct DiskOnChip *this)
{
	int floor, chip;
	int numchips[MAX_FLOORS];
	int ret = 1;
	
	this->numchips = 0;
	this->mfr = 0;
	this->id = 0;
	
	/* For each floor, find the number of valid chips it contains */
	for (floor = 0 ; floor < MAX_FLOORS ; floor++) {
		ret = 1;
		numchips[floor]=0;
		for (chip = 0 ; chip < MAX_CHIPS && ret != 0; chip++ ) {
			
			ret = DoC_IdentChip(this, floor, chip);
			if (ret) {
				numchips[floor]++;
				this->numchips++;
			}
		}
	}
	
	/* If there are none at all that we recognise, bail */
	if (!this->numchips) {
		printk("No flash chips recognised.\n");
		return;
	}
	
	/* Allocate an array to hold the information for each chip */
	this->chips = kmalloc(sizeof(struct Nand) * this->numchips, GFP_KERNEL);
	if (!this->chips){
		printk("No memory for allocating chip info structures\n");
		return;
	}
	
	ret = 0;
	
	/* Fill out the chip array with {floor, chipno} for each 
	 * detected chip in the device. */
	for (floor = 0; floor < MAX_FLOORS; floor++) {
		for (chip = 0 ; chip < numchips[floor] ; chip++) {
			this->chips[ret].floor = floor;
			this->chips[ret].chip = chip;
			this->chips[ret].curadr = 0;
			this->chips[ret].curmode = 0x50;
			ret++;
		}
	}

	/* Calculate and print the total size of the device */
	this->totlen = this->numchips * (1 << this->chipshift);

	printk(KERN_INFO "%d flash chips found. Total DiskOnChip size: %ld Mb\n", this->numchips ,
	       this->totlen >> 20);
}

/* doccheck: Probe a given memory window to see if there's a DiskOnChip present */

static inline int doccheck(unsigned long potential, unsigned long physadr)
{
	unsigned long window=potential;
	unsigned char tmp, ChipID;
#ifndef DOC_PASSIVE_PROBE
	unsigned char tmp2;
#endif

	/* Routine copied from the Linux DOC driver */
	
	/* Check for 0x55 0xAA signature at beginning of window */
	if (readb(window) != 0x55 || readb(window+1) != 0xaa)
		return 0;

#ifndef DOC_PASSIVE_PROBE	
	/* It's not possible to cleanly detect the DiskOnChip - the
	 * bootup procedure will put the device into reset mode, and
	 * it's not possible to talk to it without actually writing
	 * to the DOCControl register. So we store the current contents
	 * of the DOCControl register's location, in case we later decide
	 * that it's not a DiskOnChip, and want to put it back how we
	 * found it. 
	 */
	tmp2 = ReadDOC(window, DOCControl);
	
	/* Reset the DiskOnChip ASIC */
	WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_RESET, 
		 window, DOCControl);
	WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_RESET, 
		 window, DOCControl);
	
	/* Enable the DiskOnChip ASIC */
	WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_NORMAL, 
		 window, DOCControl);
	WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_NORMAL, 
		 window, DOCControl);
#endif /* !DOC_PASSIVE_PROBE */	

	ChipID = ReadDOC(window, ChipID);
  
	switch (ChipID) {
	case DOC_ChipID_Doc2k:
		printk(KERN_NOTICE "DiskOnChip 2000 found at address 0x%lX\n",physadr);
		/* Check the TOGGLE bit in the ECC register */
		tmp = ReadDOC(window, 2k_ECCStatus) & DOC_TOGGLE_BIT;
		if ((ReadDOC(window, 2k_ECCStatus) & DOC_TOGGLE_BIT) != tmp)
				return ChipID;
		break;
		
	case DOC_ChipID_DocMil:
		printk(KERN_NOTICE "DiskOnChip Millennium found at address 0x%lX\n",physadr);
		/* We don't actually handle the MDOC yet - it's all hard-coded for the DOC2k.
		   As M-Systems have now provided me with a sample MDOC, I'll fix this as soon
		   as I've got DOC2k support completed. 
		*/

		/* Check the TOGGLE bit in the ECC register */
		tmp = ReadDOC(window, ECCConf) & DOC_TOGGLE_BIT;
		if ((ReadDOC(window, ECCConf) & DOC_TOGGLE_BIT) != tmp)
				return ChipID;
		break;
		
	default:
		printk(KERN_WARNING "DiskOnChip with unknown ChipID %2.2X found at 0x%lx\n",
		       ChipID, physadr);

#ifndef DOC_PASSIVE_PROBE
		/* Put back the contents of the DOCControl register, in case it's not
		 * actually a DiskOnChip.
		 */
		WriteDOC(tmp2, window, DOCControl);
#endif
		return 0;
	}

	printk(KERN_WARNING "DiskOnChip failed TOGGLE test, dropping.\n");

#ifndef DOC_PASSIVE_PROBE
	/* Put back the contents of the DOCControl register: it's not a DiskOnChip */
	WriteDOC(tmp2, window, DOCControl);
#endif
	return 0;
}   

/* DoC_Probe: Probe for a DiskOnChip in a given memory window, set up and register
   an MTD device for it if neccesary */

void DoC_Probe(unsigned long physadr)
{
	unsigned long docptr = (unsigned long)ioremap(physadr, 0x2000);
	struct DiskOnChip *this;
	struct mtd_info *mtd;
	int ChipID;
	
	if (!docptr) return;
	
	if ((ChipID = doccheck(docptr, physadr))) {
		
		mtd = kmalloc(sizeof(struct DiskOnChip) + sizeof(struct mtd_info), GFP_KERNEL);

		if (!mtd) {
			printk("Cannot allocate memory for data structures. Dropping.\n");
			iounmap((void *)docptr);
			return;
		}
		
		this = (struct DiskOnChip *)(&mtd[1]);
		
		memset((char *)mtd,0, sizeof(struct mtd_info));
		
		mtd->priv = this;
		
		switch(ChipID) {
		case DOC_ChipID_Doc2k:
			strcpy(mtd->name, "DiskOnChip 2000");
			break;
			
		case DOC_ChipID_DocMil:
			strcpy(mtd->name, "DiskOnChip Millennium");
			break;
			
		default:
			strcpy(mtd->name, "DiskOnChip");
		}
		mtd->type = MTD_NANDFLASH;
		mtd->flags = MTD_CAP_NANDFLASH;
		mtd->size = 0;
		mtd->erasesize = 0x2000;
		mtd->oobblock = 512;
		mtd->oobsize = 16;
		mtd->module = THIS_MODULE;
		mtd->erase = doc_erase;
		mtd->point = NULL;
		mtd->unpoint = NULL;
		mtd->read = doc_read;
		mtd->write = doc_write;
		mtd->read_ecc = doc_read_ecc;
		mtd->write_ecc = doc_write_ecc;
		mtd->read_oob = doc_read_oob;
		mtd->write_oob = doc_write_oob;
		mtd->sync = NULL;

		this->physadr = physadr;
		this->virtadr = docptr;
		this->ChipID = ChipID;
		
		this->totlen = 0;
		this->numchips = 0;
		
		this->curfloor = -1;
		this->curchip = -1;
		
		/* Ident all the chips present. */
		DoC_ScanChips(this);

		if (!this->totlen)
			kfree(mtd);
		else {
			numdocs++;
			mtd->size  = this->totlen;
			add_mtd_device(mtd);
			return;
		}
	}
	
	/* We failed to detect a valid device. Unmap the address range */
	iounmap((void *)docptr);
}


static int doc_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf)
{
	/* Just a special case of doc_read_ecc */
	return doc_read_ecc(mtd, from, len, retlen, buf, NULL);
}

static int doc_read_ecc (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf, u_char *eccbuf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int di=0; /* Yes, DI is a hangover from when I was disassembling the binary driver */
	unsigned long docptr;
	struct Nand *mychip;

	docptr = this->virtadr;

	/* Don't allow read past end of device */
	if (from >= this->totlen)
		return -EINVAL;
	
	/* Don't allow a single read to cross a 512-byte block boundary */
	if (from + len > ( (from | 0x1ff) + 1)) 
		len = ((from | 0x1ff) + 1) - from;

	/* Find the chip which is to be used and select it */
	mychip = &this->chips[from >> (this->chipshift)];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	

	if (eccbuf) {
		/* Prime the ECC engine */
		WriteDOC ( DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC ( DOC_ECC_EN, docptr, ECCConf);
	}

	DoC_Command(docptr, (from >> 8) & 1, CSDN_CTRL_WP);
	DoC_Address(docptr, 3, from, CSDN_CTRL_WP , CSDN_CTRL_ECC_IO);

	for (di=0; di < len ; di++) {
		buf[di] = ReadDOC(docptr, 2k_CSDN_IO);
	}

	/* Let the caller know we completed it */
	*retlen = len;

	if (eccbuf) {
		/* Read the ECC data through the DiskOnChip ECC logic */
		for (di=0; di<6; di++) {
			eccbuf[di] = ReadDOC(docptr, 2k_CSDN_IO);
		}
		
		/* Flush the pipeline */
		(void) ReadDOC(docptr, 2k_ECCStatus);
		(void) ReadDOC(docptr, 2k_ECCStatus);
		
		/* Check the ECC Status */
		if (ReadDOC(docptr, 2k_ECCStatus) & 0x80) {
			/* There was an ECC error */
			printk("DiskOnChip ECC Error: Read at %lx\n", (long)from);

			/* FIXME: Implement ECC error correction, don't just whinge */

			/* We return error, but have actually done the read. Not that
			   this can be told to user-space, via sys_read(), but at least
			   MTD-aware stuff can know about it by checking *retlen */
			return -EIO;
		}
#ifdef PSYCHO_DEBUG
		else
			printk("ECC OK at %lx: %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			       (long)from, eccbuf[0], eccbuf[1], eccbuf[2], eccbuf[3], eccbuf[4],
			       eccbuf[5]);
#endif
		
		/* Reset the ECC engine */
		WriteDOC(DOC_ECC_RESV, docptr , ECCConf);
		
	}

	return 0;
}

static int doc_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf)
{
	static char as[6];
	return doc_write_ecc(mtd, to, len, retlen, buf, as);
}

static int doc_write_ecc (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf, u_char *eccbuf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int di=0; 
	unsigned long docptr;
	struct Nand *mychip;

	docptr = this->virtadr;

	/* Don't allow write past end of device */
	if (to >= this->totlen)
		return -EINVAL;
#if 0	
	/* Don't allow a single write to cross a 512-byte block boundary */
	if (to + len > ( (to | 0x1ff) + 1)) 
		len = ((to | 0x1ff) + 1) - to;

#else
	/* Don't allow writes which aren't exactly one block */
	if (to & 0x1ff || len != 0x200)
		return -EINVAL;
#endif

	/* Find the chip which is to be used and select it */
	mychip = &this->chips[to >> (this->chipshift)];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	/* Set device to main plane of flash */
	DoC_Command(docptr, NAND_CMD_RESET, CSDN_CTRL_WP);
	DoC_Command(docptr, NAND_CMD_READ0, CSDN_CTRL_WP);

	if (eccbuf) {
		/* Prime the ECC engine */
		WriteDOC ( DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC ( DOC_ECC_EN | DOC_ECC_RW, docptr, ECCConf);
	}

	DoC_Command(docptr, NAND_CMD_SEQIN, 0);
	DoC_Address(docptr, 3, to, 0, CSDN_CTRL_ECC_IO);

	for (di=0; di < len ; di++) {
		WriteDOC(buf[di], docptr, 2k_CSDN_IO);
	}


	if (eccbuf) {
		WriteDOC( CSDN_CTRL_ECC_IO | CSDN_CTRL_CE , docptr, CSDNControl );
		
#if 1
		/* eduardp@m-sys.com says this shouldn't be necessary,
		 * but it doesn't actually work without it, so I've
		 * left it in for now. dwmw2.
		 */
		 
		WriteDOC( 0, docptr, 2k_CSDN_IO);
		WriteDOC( 0, docptr, 2k_CSDN_IO);
		WriteDOC( 0, docptr, 2k_CSDN_IO);
#endif
		/* Read the ECC data through the DiskOnChip ECC logic */
		for (di=0; di<6; di++) {
			eccbuf[di] = ReadDOC(docptr, ECCSyndrome0 + di);
		}
#ifdef PSYCHO_DEBUG
		printk("OOB data at %lx is %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
		       (long) to, eccbuf[0], eccbuf[1], eccbuf[2],
		       eccbuf[3], eccbuf[4], eccbuf[5]       );
#endif
		/* Reset the ECC engine */
		WriteDOC(DOC_ECC_RESV, docptr , ECCConf);
		
	}

	DoC_Command(docptr, NAND_CMD_PAGEPROG, 0);

	DoC_Command(docptr, NAND_CMD_STATUS, CSDN_CTRL_WP);
	/* There's an implicit DoC_WaitReady() in DoC_Command */

	if (ReadDOC(docptr, 2k_CSDN_IO) & 1) {
		printk("Error programming flash\n");
		/* Error in programming */
		*retlen = 0;
		return -EIO;
	}

	/* Let the caller know we completed it */
	*retlen = len;

	return 0;
}



static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int i;
	unsigned long docptr;
	struct Nand *mychip;
	
	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	
	
	DoC_Command(docptr, NAND_CMD_READOOB, CSDN_CTRL_WP);
	DoC_Address(docptr, 3, ofs, CSDN_CTRL_WP, 0);
	
	for (i=0; i<len; i++)
		buf[i] = ReadDOC(docptr, 2k_CSDN_IO);
	
	*retlen = len;
	return 0;

}

static int doc_write_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, const u_char *buf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int i;
	unsigned long docptr;
	struct Nand *mychip;

	//	printk("doc_write_oob(%lx, %d): %2.2X %2.2X %2.2X %2.2X ... %2.2X %2.2X .. %2.2X %2.2X\n",(long)ofs, len,
	//   buf[0], buf[1], buf[2], buf[3], buf[8], buf[9], buf[14],buf[15]);

	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	DoC_Command(docptr, NAND_CMD_RESET, CSDN_CTRL_WP);
	DoC_Command(docptr, NAND_CMD_READOOB, CSDN_CTRL_WP);

	DoC_Command(docptr, NAND_CMD_SEQIN, 0);
	DoC_Address(docptr, 3, ofs, 0, 0);
	
	for (i=0; i<len; i++)
		WriteDOC(buf[i], docptr, 2k_CSDN_IO);

	DoC_Command(docptr, NAND_CMD_PAGEPROG, 0);
	DoC_Command(docptr, NAND_CMD_STATUS, 0);
	/* DoC_WaitReady() is implicit in DoC_Command */

	if (ReadDOC(docptr, 2k_CSDN_IO) & 1) {
		printk("Error programming oob data\n");
		/* There was an error */
		*retlen = 0;
		return -EIO;
	}

	*retlen = len;
	return 0;

}


int doc_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	unsigned long ofs = instr->addr;
	unsigned long len = instr->len;
	unsigned long docptr;
	struct Nand *mychip;
	
	if(len != mtd->erasesize) 
		printk(KERN_WARNING "Erase not right size (%lx != %lx)n", len, mtd->erasesize);
		

	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	DoC_Command(docptr, NAND_CMD_ERASE1, 0);
	DoC_Address(docptr, 2, ofs, 0, 0);
	DoC_Command(docptr, NAND_CMD_ERASE2, 0);

	DoC_Command(docptr, NAND_CMD_STATUS, CSDN_CTRL_WP);

	if (ReadDOC(docptr, 2k_CSDN_IO) & 1) {
		printk("Error writing\n");
		/* There was an error */
		kfree(instr);
		return -EIO;
	}
	kfree(instr);
	return 0;
}


/****************************************************************************
 *
 * Module stuff
 *
 ****************************************************************************/

#if LINUX_VERSION_CODE < 0x20300
#ifdef MODULE
#define init_doc2000 init_module
#define cleanup_doc2000 cleanup_module
#endif
#define __exit
#endif

static int __init init_doc2000(void)
{
	int i;
	
	
	printk(KERN_NOTICE "M-Systems DiskOnChip driver. (C) 1999 Machine Vision Holdings, Inc.\n");
#ifdef PRERELEASE
	printk(KERN_INFO "$Id: doc2000.c,v 1.9 2000/01/19 10:26:04 dwmw2 Exp $\n");
#endif
	
	for (i=0; doc_locations[i]; i++) {
		DoC_Probe(doc_locations[i]);
	}
	
	if (numdocs)
		return 0;
	
	printk("No DiskOnChip hardware found\n");
	return -ENXIO;
	
}

static void __exit cleanup_doc2000(void)
{
	int i;
	struct mtd_info *mtd;
	
	for (i=0; i < MAX_MTD_DEVICES; i++) {
		mtd = get_mtd_device(i);
		
		if (mtd && mtd->module == THIS_MODULE) {
			struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
			
			del_mtd_device(mtd);
			
			iounmap((void *)this->virtadr);
			kfree(this->chips);
			kfree(mtd);
		}
	}
}

