// -*- mode: cpp; mode: fold -*-
// Description                                                          /*{{{*/
// $Id: mtdblock.c,v 1.4 2000/01/19 10:26:04 dwmw2 Exp $
/* ######################################################################

   Memory Technology Devices Manager

   TODO:
     - Add a ioctl method to the mtd structure and move all flash related
       stuff into a new separated structure so as not to pollute the 
       main mtd info.
     - Flexible partitioning
   
   ##################################################################### */
									/*}}}*/
#include <linux/types.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/config.h>
#include <linux/init.h>

#include <linux/mtd/mtd.h>

#define MAJOR_NR MTD_BLOCK_MAJOR
#define DEVICE_NAME "mtdblock"
#define DEVICE_REQUEST mtdblock_request
#define DEVICE_NR(device) (device)
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#include <linux/blk.h>

static int mtd_blksizes[MAX_MTD_DEVICES];
static int mtd_sizes[MAX_MTD_DEVICES];


// mtd_open - Open a block device					/*{{{*/
// ---------------------------------------------------------------------
/* Checks that someone has been assigned that device number. */
static int mtdblock_open(struct inode *inode, struct file *file)
{
	struct mtd_info *mtd = NULL;

	int dev;


	printk("mtdblock_open\n");

	if (inode == 0)
		return -EINVAL;
	
	dev = MINOR(inode->i_rdev);
	
	mtd = get_mtd_device(dev);

	if (!mtd)
		return -ENODEV;

	MOD_INC_USE_COUNT;

	if (mtd->module)
		__MOD_INC_USE_COUNT(mtd->module);
	printk("ok\n");
	return 0;
}
									/*}}}*/
// mtd_release - Close a mtd file					/*{{{*/
// ---------------------------------------------------------------------
/* Flushes the buffers on the device when closed and makes sure it is 
   still registered */
static int mtdblock_release(struct inode *inode, struct file *file)
{
	int dev;
	struct mtd_info *mtd;

   	printk("mtdblock_release\n");

	if (inode == NULL)
		return -ENODEV;
   
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);

	dev = MINOR(inode->i_rdev);
	mtd = get_mtd_device(dev);

	if (!mtd) {
		printk(KERN_WARNING "MTD device is absent on mtd_release!\n");
		MOD_DEC_USE_COUNT;
		return -ENODEV;
		
	}

	if (mtd->module)
		__MOD_DEC_USE_COUNT(mtd->module);
	printk("ok\n");
	MOD_DEC_USE_COUNT;
	return 0;
}  


// mtd_request - Read/Write request					/*{{{*/
// ---------------------------------------------------------------------
/* This is a simple request handler to verify the parameters and pass them
   on to the lower mtd layer */
static void mtdblock_request(void)
{
   struct request *current_request;
   unsigned int res = 0;
   struct mtd_info *mtd;

   while (1)
   {
      /* Grab the Request and unlink it from the request list, INIT_REQUEST
       	 will execute a return if we are done. */
      INIT_REQUEST;
      current_request = CURRENT;
   
      if (MINOR(current_request->rq_dev) >= MAX_MTD_DEVICES)
      {
	 printk("mtd: Unsupported device!\n");
	 end_request(0);
	 continue;
      }
      
      // Grab our MTD structure
      mtd = get_mtd_device(CURRENT_DEV);
      if (mtd == 0 || current_request->sector << 9 > mtd->size ||
	  (current_request->sector + current_request->nr_sectors) << 9 > mtd->size)
      {
	 printk("mtd: Attempt to read past end of device!\n");
	 end_request(0);
	 continue;
      }
      
      // Grab the device lock
      if (mtd->lock != 0 && test_and_set_bit(0,mtd->lock))
      {      
	 printk("mtd: request while busy, deferred\n");
	 
	 // Returning should delay a bit before coming back I hope
	 return;
      }
      
      /* Remove the request we are handling from the request list so nobody messes
         with it */
      CURRENT = current_request->next;
      
      /* Now drop the lock that the ll_rw_blk functions grabbed for us
         and process the request. This is necessary due to the extreme time
         we spend processing it. */
      spin_unlock_irq(&io_request_lock);
      
      // Handle the request
      switch (current_request->cmd)
      {
         size_t retlen;

	 case READ:
	 if (MTD_READ(mtd,current_request->sector<<9, 
		      current_request->nr_sectors << 9, 
		      &retlen, current_request->buffer) == 0)
	    res = 1;
	 else
	    res = 0;
	 break;
	 
	 case WRITE:
	 // Read only device
	 if ((mtd->flags & MTD_CAP_RAM) == 0)
	 {
	    res = 0;
	    break;
	 }
	 
	 // Do the write
	 if (MTD_WRITE(mtd,current_request->sector<<9, 
		       current_request->nr_sectors << 9, 
		       &retlen, current_request->buffer) == 0)
	    res = 1;
	 else
	    res = 0;
	 break;
	 
	 // Shouldn't happen
	 default:
	 printk("mtd: unknown request\n");
	 break;
      }
      
      if (mtd->lock != 0)
	 clear_bit(0,mtd->lock);
      
      // Grab the lock and re-thread the item onto the linked list
      spin_lock_irq(&io_request_lock);
      current_request->next = CURRENT;
      CURRENT = current_request;
      end_request(res);
   }
}
									/*}}}*/

static struct file_operations mtd_fops =
{
	NULL,	             /* lseek - default */
	block_read,	     /* read - general block-dev read */
	block_write,	     /* write - general block-dev write */
	NULL,		     /* readdir - bad */
	NULL,		     /* select */
	NULL,     	     /* ioctl */
	NULL,		     /* mmap */
	mtdblock_open,	     /* open */
	NULL,		     /* flush */
	mtdblock_release,    /* release */
	block_fsync          /* fsync */  
};




#if LINUX_VERSION_CODE < 0x20300
#ifdef MODULE
#define init_mtdblock init_module
#define cleanup_mtdblock cleanup_module
#endif
#define __exit
#endif


// init_module - Module entry point					/*{{{*/
// ---------------------------------------------------------------------
/* */
static int __init init_mtdblock(void)
{

   if (register_blkdev(MAJOR_NR,DEVICE_NAME,&mtd_fops))
   {
      printk(KERN_NOTICE "Can't allocate major number %d for Memory Technology Devices.\n",
	     MTD_BLOCK_MAJOR);
      return EAGAIN;
   }

   printk("MAJOR_NR == %d, MTD_BLOCK_MAJOR %d\n",MAJOR_NR, MTD_BLOCK_MAJOR);
   blksize_size[MAJOR_NR] = mtd_blksizes;
   blk_size[MAJOR_NR] = mtd_sizes;
   blk_dev[MAJOR_NR].request_fn = mtdblock_request;
   
   return 0;
}
									/*}}}*/
// cleanup_module - Module exit point					/*{{{*/
// ---------------------------------------------------------------------
/* */

static void __exit cleanup_mtdblock(void)
{
   unregister_blkdev(MAJOR_NR,DEVICE_NAME);
}

#if LINUX_VERSION_CODE > 0x20300
module_init(init_mtdblock);
module_exit(cleanup_mtdblock);
#endif
