// -*- mode: cpp; mode: fold -*-
// Description                                                          /*{{{*/
// $Id: vmax301.c,v 1.4 1999/10/25 14:55:46 dwmw2 Exp $
/* ######################################################################

   Tempustech VMAX SBC301 MTD Driver.
  
   The VMAx 301 is a SBC based on . It
   comes with three builtin AMD 29F016B flash chips and a socket for SRAM or
   more flash. Each unit has it's own 8k mapping into a settable region 
   (0xD8000). There are two 8k mappings for each MTD, the first is always set
   to the lower 8k of the device the second is paged. Writing a 16 bit page
   value to anywhere in the first 8k will cause the second 8k to page around.

   To boot the device a bios extension must be installed into the first 8k 
   of flash that is smart enough to copy itself down, page in the rest of 
   itself and begin executing.
   
   ##################################################################### */
									/*}}} */
// Include Files                                                        /*{{{*/
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <asm/io.h>

#include <linux/mtd/mapped.h>
									/*}}} */

#define WINDOW_START 0xd8000
#define WINDOW_LENGTH 0x2000

static struct mapped_mtd_info *SSD[2];
static int VMaxLock = 0;

// vmax301_page - Return a readb address for the new page window	/*{{{*/
// ---------------------------------------------------------------------
/* All page windows are of the page window size set at map creation, 
   map->pagesize. All this function does is move the page window and 
   return the address */
static unsigned long vmax301_page(struct mapped_mtd_info *map,
				  unsigned long page)
{
   unsigned long Device = (unsigned long)map->mtd.priv;
   writew(page,WINDOW_START + 2*Device*WINDOW_LENGTH);
   return WINDOW_START + WINDOW_LENGTH + 2*Device*WINDOW_LENGTH;
}
									/*}}}*/
									/*}}}*/

// CreateSSD - Construct the SSD structure				/*{{{*/
// ---------------------------------------------------------------------
/* Since two SSDs are created this simplifies the init code.. */
static struct mapped_mtd_info *CreateSSD()
{
   // Allocate some memory
   struct mapped_mtd_info *mtd;
   mtd = (struct mapped_mtd_info *)kmalloc(sizeof(*mtd),GFP_KERNEL);
   if (mtd == 0)
      return 0;   
   memset(mtd,0,sizeof(*mtd));

   // Setup the MTD structure
   mtd->page = vmax301_page;
   mtd->pagesize = WINDOW_LENGTH;
   mtd->maxsize = 0xFFFF;
#ifdef MODULE
   mtd->mtd.module = &__this_module;
#endif
   /* Give it a lock for our driver instance. This will prevent ANY 
      concurrent activity between the two flash devices using the window */
   mtd->mtd.lock = &VMaxLock;
   
   return mtd;
}
									/*}}}*/

#if LINUX_VERSION_CODE < 0x20300
#ifdef MODULE
#define init_vmax301 init_module
#define cleanup_vmax301 cleanup_module
#endif
#define __exit
#endif


// cleanup_module - Clean up the allocations made			/*{{{*/
// ---------------------------------------------------------------------
/* */
static void __exit cleanup_vmax301()
{
   int I;
   for (I = 0; I != 2; I++)
   {
      mtd_mapped_remove(SSD[I]);
      kfree(SSD[I]);
      SSD[I] = 0;
   }
}
									/*}}}*/
// init_module - Module startup function				/*{{{*/
// ---------------------------------------------------------------------
/* Create the two SSD devices */
static int __init init_vmax301()
{
   // Print out our little header..
   printk("Tempustech VMAX 301 MEM:0x%x-0x%x\n",WINDOW_START,
	  WINDOW_START+4*WINDOW_LENGTH);
   	  
   // Create the SSDs
   SSD[0] = CreateSSD();
   SSD[1] = CreateSSD();
   if (SSD[0] == 0 || SSD[1] == 0)
   {
      cleanup_vmax301();
      return -EAGAIN;
   }

   /* We set characteristics for two ROM devices of 0 and 6M, I don't know
      how to sense ROM size yet.. */
   strcpy(SSD[0]->mtd.name,"VMAX Socket");
   SSD[0]->mtd.size = 0;
   SSD[0]->mtd.type = MTD_ROM;
   (unsigned long)SSD[0]->mtd.priv = 1;
   strcpy(SSD[1]->mtd.name,"VMAX Internal");
   SSD[1]->mtd.size = 3*2*1024*1024;
   SSD[1]->mtd.type = MTD_ROM;
   (unsigned long)SSD[1]->mtd.priv = 0;
   
   // If either are flash then the correct settings will be updated now..
   if (mtd_mapped_setup(SSD[0]) != 0 || mtd_mapped_setup(SSD[1]) != 0)
   {
      printk("Failed to register new device\n");
      cleanup_vmax301();
      return -EAGAIN;
   }
   
   return 0;
}
									/*}}}*/
#if LINUX_VERSION_CODE > 0x20300
module_init(init_vmax301);
module_exit(cleanup_vmax301);
#endif
